# Please maintain compatibility with Python 2.7 and 3+.

import csv
from datetime import date, datetime, timedelta
from time import sleep
import json
import os
import re
import requests
import sys

AD_INSIGHTS_TOKEN = os.environ['AD_INSIGHTS_TOKEN']
API_ROOT_URL = 'https://graph.facebook.com/'
API_BASE_URL = 'https://graph.facebook.com/v9.0/'
API_BASE_URL = 'https://graph.facebook.com/v10.0/'

def stdin_lines():
    while True:
        line = sys.stdin.readline()
        if not line:
            break
        yield line

def stdin_words():
    for line in stdin_lines():
        if line.strip():
            for word in line.split():
                yield word

def out(message, *args):
    sys.stdout.write((message % args) if args else message)
    sys.stdout.flush()

def warn(message, *args):
    sys.stderr.write('\x1b[33m')
    log(message, *args)

def err(message, *args):
    sys.stderr.write('\x1b[31m')
    log(message, *args)

def yay(message, *args):
    sys.stderr.write('\x1b[32m')
    log(message, *args)

def debug(message, *args):
    if os.environ.get('DEBUG'):
        sys.stderr.write('\x1b[34m')
        log(message, *args)

def log(message, *args):
    sys.stderr.write((message % args) if args else message)
    sys.stderr.write('\x1b[0m')
    sys.stderr.flush()

def truncate(str, maxlen):
    if len(str) > maxlen:
        return str[:maxlen] + '...'
    return str

# Converts lists of pairs to a set of columns; this is intended to
# handle the fields of type list<AdsActionStats>.
def ads_pairs_to_csv(key, pairs, name_key, value_key):
    return {'%s_%s' % (key, pair.get(name_key)): pair.get(value_key)
        for pair in pairs}

# Converts lists of scalars to a set of columns; this is intended to
# handle the fields of type list<AdsHistogramStats>.
def ads_array_to_csv(key, array):
    return {'key_%d' % (key, i): item for i, item in enumerate(array)}

def json_object_to_csv(json_object):
    row = {}
    for key, value in json_object.items():
        if isinstance(value, list):
            if value and isinstance(value[0], dict):
                row.update(ads_pairs_to_csv(key, value, 'action_type', 'value'))
            else:
                row.update(ads_array_to_csv(key, value))
        else:
            row[key] = value
    return row

def parse_json(text):
    try:
        return json.loads(text)
    except ValueError:
        err('invalid JSON: %r\n' % truncate(text, 60))
        return {}

def count(n, singular='', plural=None):
    if not isinstance(n, (int, float)):
        n = len(n)
    if n == 1:
        return '%d %s' % (n, singular)
    else:
        return '%d %s' % (n, plural or singular + 's')

def read(path):
    if isinstance(path, str):
        with open(path + '.tmp', 'wb') as file:
            return file.read().decode('utf-8')
    else:
        return path.read().decode('utf-8')

def read_json(path):
    return parse_json(read(path))

def read_jsonl(path):
    if isinstance(path, str):
        with open(path + '.tmp', 'wb') as file:
            line = file.readline()
            while line:
                yield parse_json(line.decode('utf-8'))
                line = file.readline()
    else:
        line = path.readline()
        while line:
            yield parse_json(line.decode('utf-8'))
            line = path.readline()

def read_normalized_jsonl(path):
    return normalize_jsons_for_bigquery(read_jsonl(path))

def write(path, text):
    if isinstance(path, str):
        with open(path + '.tmp', 'wb') as file:
            file.write(text.encode('utf-8'))
        os.rename(path + '.tmp', path)
        yay('=> %s (%s)\n' % (path, count(text, 'char')))
    else:
        path.write(text.encode('utf-8'))
        yay('=> %r (%s)\n' % (path, count(text, 'char')))

def write_items(path, items):
    length = 0
    if isinstance(path, str):
        with open(path + '.tmp', 'wb') as file:
            for item in items:
                file.write(str(item) + '\n')
                file.flush()
                length += 1
        os.rename(path + '.tmp', path)
        yay('=> %s (%s)\n' % (path, count(length, 'item')))
    else:
        for item in items:
            path.write(str(item) + '\n')
            path.flush()
            length += 1
        yay('=> %r (%s)\n' % (path, count(length, 'item')))

def write_json(path, obj):
    write(path, json.dumps(obj, indent=2, sort_keys=1))

def write_jsonl(path, objs):
    length = 0
    if isinstance(path, str):
        with open(path + '.tmp', 'wb') as file:
            for obj in objs:
                file.write(json.dumps(obj, sort_keys=1).encode('utf-8') + b'\n')
                file.flush()
                length += 1
        os.rename(path + '.tmp', path)
        yay('=> %s (%s)\n' % (path, count(length, 'object')))
    else:
        for obj in objs:
            path.write(json.dumps(obj, sort_keys=1).encode('utf-8') + b'\n')
            path.flush()
            length += 1
        yay('=> %r (%s)\n' % (path, count(length, 'object')))

def write_normalized_jsonl(path, objs):
    write_jsonl(path, normalize_jsons_for_bigquery(objs))

def write_csv(path, rows):
    fieldnames = set()
    rows = list(rows)
    for row in rows:
        fieldnames.update(row.keys())
    if isinstance(path, str):
        try:
            with open(path + '.tmp', 'w') as file:
                writer = csv.DictWriter(file, sorted(fieldnames))
                if rows:
                    writer.writeheader()
                for row in rows:
                    writer.writerow([field.encode('utf-8') for field in rows])
                file.flush()
        except Exception as e:
            err('problem writing to %s: %s' % (path, e))
        else:
            os.rename(path + '.tmp', path)
            yay('=> %s (%s)\n' % (path, count(rows, 'row')))
    else:
        writer = csv.DictWriter(path, sorted(fieldnames))
        if rows:
            writer.writeheader()
        for row in rows:
            writer.writerow([field.encode('utf-8') for field in rows])
            path.flush()
        yay('=> %r (%s)\n' % (path, count(rows, 'row')))

def normalize_json_for_bigquery(obj):
    if isinstance(obj, dict):
        result = {}
        for key in sorted(obj.keys()):
            norm_key = get_normalized_key(key, result)
            result[norm_key] = normalize_json_for_bigquery(obj[key])
        return result
    if isinstance(obj, list):
        return [normalize_json_for_bigquery(item) for item in flatten_lists(obj)]
    return obj

def normalize_jsons_for_bigquery(objs):
    for obj in objs:
        yield normalize_json_for_bigquery(obj)

def flatten_lists(items):
    result = []
    for item in items:
        if isinstance(item, list):
            result += flatten_lists(item)
        else:
            result.append(item)
    return result

def get_normalized_key(key, obj={}):
    norm = re.sub(r'\W', '_', key)
    if re.match(r'^\d', norm):  # BigQuery rejects keys that start with a digit
        norm = '_' + norm
    if norm == key:
        return key
    candidate, suffix = norm, 1
    while candidate in obj:
        suffix += 1
        candidate = '%s_%d' % (nkey, suffix)
    return candidate

def get_api_url(path):
    if re.match(r'^https?://', path):
        return path
    if re.match(r'^/', path):
        return API_ROOT_URL + path.lstrip('/')
    return API_BASE_URL + path.lstrip('/')

def fetch(method, url, **params):
    debug(url + ' - ' + repr(params) + '\n')
    return method(url, **params)

def get_text(path, **params):
    params['access_token'] = AD_INSIGHTS_TOKEN
    response = fetch(requests.get, get_api_url(path), params=params)
    comply_with_rate_limits(response)
    return response.text

def get(path, **params):
    params['access_token'] = AD_INSIGHTS_TOKEN
    response = fetch(requests.get, get_api_url(path), params=params)
    comply_with_rate_limits(response)
    return parse_json(response.text)

def post(path, **params):
    params['access_token'] = AD_INSIGHTS_TOKEN
    response = fetch(requests.post, get_api_url(path), data=params)
    comply_with_rate_limits(response)
    return parse_json(response.text)

def get_pages(path, **params):
    next_url = path
    while next_url:
        data = get(next_url, **params)
        if 'error' in data:
            raise RuntimeError(data)
        yield data
        next_url = data.get('paging', {}).get('next')
        params = {}

def get_records(path, **params):
    for page in get_pages(path, **params):
        log('.')
        for record in page.get('data', []):
            yield record
    log('\n')

def comply_with_rate_limits(response):
    try:
        content = response.json()
    except:
        content = {}
    if isinstance(content, str):
        content = {}
    error = content.get('error') or {}
    message = error.get('message') or ''
    if error:
        if error.get('code') == 80004 or 'too many calls' in message:
            err('\n' + message)
            warn('\n[utilization is too high; pausing]\n')
            sleep(60)
            yay('\n[pause completed; continuing]\n')

    throttle = response.headers.get('X-FB-Ads-Insights-Throttle',
        response.headers.get('X-Ad-Account-Usage')
    )
    if throttle:
        utilization = 0
        try:
            values = json.loads(throttle)
            utilization = float(values.get('acc_id_util_pct'))
            warn('\n%s: [utilization: %d%%]\n' % (
                datetime.utcnow().isoformat().split('.')[0], utilization))
        except:
            utilization = 0
        if utilization > 70:
            warn('\n[utilization is too high; pausing]\n')
            sleep(60)
            yay('\n[pause completed; continuing]\n')

if __name__ == '__main__':
    main(sys.argv[1:])
